@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 2019-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 2019-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -25,7 +25,6 @@
 #include "nv-pci-types.h"
 #include "nv-pci.h"
 #include "nv-ibmnpu.h"
-#include "nv-frontend.h"
 #include "nv-msi.h"
 #include "nv-hypervisor.h"
 
@@ -80,19 +79,15 @@ static NvBool nv_treat_missing_irq_as_error(void)
 #endif
 }
 
-static void nv_init_dynamic_power_management
+static void nv_get_pci_sysfs_config
 (
-    nvidia_stack_t *sp,
-    struct pci_dev *pci_dev
+    struct pci_dev *pci_dev,
+    nv_linux_state_t *nvl
 )
 {
-    nv_linux_state_t *nvl = pci_get_drvdata(pci_dev);
-    nv_state_t *nv = NV_STATE_PTR(nvl);
+#if NV_FILESYSTEM_ACCESS_AVAILABLE
     char filename[50];
     int ret;
-    NvBool pr3_acpi_method_present = NV_FALSE;
-
-    nvl->sysfs_config_file = NULL;
 
     ret = snprintf(filename, sizeof(filename),
                    "/sys/bus/pci/devices/%04x:%02x:%02x.0/config",
@@ -143,6 +138,22 @@ static void nv_init_dynamic_power_management
 #endif
         }
     }
+#endif
+}
+
+static void nv_init_dynamic_power_management
+(
+    nvidia_stack_t *sp,
+    struct pci_dev *pci_dev
+)
+{
+    nv_linux_state_t *nvl = pci_get_drvdata(pci_dev);
+    nv_state_t *nv = NV_STATE_PTR(nvl);
+    NvBool pr3_acpi_method_present = NV_FALSE;
+
+    nvl->sysfs_config_file = NULL;
+
+    nv_get_pci_sysfs_config(pci_dev, nvl);
 
     if (nv_get_hypervisor_type() != OS_HYPERVISOR_UNKNOWN)
     {
@@ -503,6 +514,14 @@ next_bar:
         goto err_zero_dev;
     }
 
+    nv->cpu_numa_node_id = dev_to_node(nvl->dev);
+
+    if (nv_linux_init_open_q(nvl) != 0)
+    {
+        NV_DEV_PRINTF(NV_DBG_ERRORS, nv, "nv_linux_init_open_q() failed!\n");
+        goto err_zero_dev;
+    }
+
     nv_printf(NV_DBG_INFO,
               "NVRM: PCI:%04x:%02x:%02x.%x (%04x:%04x): BAR0 @ 0x%llx (%lluMB)\n",
               nv->pci_info.domain, nv->pci_info.bus, nv->pci_info.slot,
@@ -524,13 +543,14 @@ next_bar:
      */
     LOCK_NV_LINUX_DEVICES();
 
-    nv_linux_add_device_locked(nvl);
+    if (nv_linux_add_device_locked(nvl) != 0)
+    {
+        UNLOCK_NV_LINUX_DEVICES();
+        goto err_add_device;
+    }
 
     UNLOCK_NV_LINUX_DEVICES();
 
-    if (nvidia_frontend_add_device((void *)&nv_fops, nvl) != 0)
-        goto err_remove_device;
-
     pm_vt_switch_required(nvl->dev, NV_TRUE);
 
     nv_init_dynamic_power_management(sp, pci_dev);
@@ -546,7 +566,6 @@ next_bar:
     if (nvidia_vgpu_vfio_probe(nvl->pci_dev) != NV_OK)
     {
         NV_DEV_PRINTF(NV_DBG_ERRORS, nv, "Failed to register device to vGPU VFIO module");
-        nvidia_frontend_remove_device((void *)&nv_fops, nvl);
         goto err_vgpu_kvm;
     }
 #endif
@@ -576,10 +595,11 @@ err_vgpu_kvm:
     nv_procfs_remove_gpu(nvl);
     rm_cleanup_dynamic_power_management(sp, nv);
     pm_vt_switch_unregister(nvl->dev);
-err_remove_device:
     LOCK_NV_LINUX_DEVICES();
     nv_linux_remove_device_locked(nvl);
     UNLOCK_NV_LINUX_DEVICES();
+err_add_device:
+    nv_linux_stop_open_q(nvl);
 err_zero_dev:
     rm_free_private_state(sp, nv);
 err_not_supported:
@@ -628,14 +648,22 @@ nv_pci_remove(struct pci_dev *pci_dev)
         return;
     }
 
-    LOCK_NV_LINUX_DEVICES();
     nvl = pci_get_drvdata(pci_dev);
     if (!nvl || (nvl->pci_dev != pci_dev))
     {
-        goto done;
+        nv_kmem_cache_free_stack(sp);
+        return;
     }
 
     nv = NV_STATE_PTR(nvl);
+
+    /*
+     * Flush and stop open_q before proceeding with removal to ensure nvl
+     * outlives all enqueued work items.
+     */
+    nv_linux_stop_open_q(nvl);
+
+    LOCK_NV_LINUX_DEVICES();
     down(&nvl->ldata_lock);
 
     /*
@@ -709,12 +737,6 @@ nv_pci_remove(struct pci_dev *pci_dev)
     nvidia_vgpu_vfio_remove(pci_dev, NV_TRUE);
 #endif
 
-    /* Update the frontend data structures */
-    if (NV_ATOMIC_READ(nvl->usage_count) == 0)
-    {
-        nvidia_frontend_remove_device((void *)&nv_fops, nvl);
-    }
-
     if ((nv->flags & NV_FLAG_PERSISTENT_SW_STATE) || (nv->flags & NV_FLAG_OPEN))
     {
         nv_acpi_unregister_notifier(nvl);
@@ -881,6 +903,21 @@ NvU8 nv_find_pci_capability(struct pci_dev *pci_dev, NvU8 capability)
     return 0;
 }
 
+static void check_for_bound_driver(struct pci_dev *pci_dev)
+{
+    if (pci_dev->dev.driver)
+    {
+        const char *driver_name = pci_dev->dev.driver->name;
+
+        nv_printf(NV_DBG_WARNINGS, "NVRM: GPU %04x:%02x:%02x.%x is already "
+            "bound to %s.\n",
+            NV_PCI_DOMAIN_NUMBER(pci_dev), NV_PCI_BUS_NUMBER(pci_dev),
+            NV_PCI_SLOT_NUMBER(pci_dev), PCI_FUNC(pci_dev->devfn),
+            driver_name ? driver_name : "another driver"
+        );
+    }
+}
+
 /* make sure the pci_driver called probe for all of our devices.
  * we've seen cases where rivafb claims the device first and our driver
  * doesn't get called.
@@ -908,6 +945,7 @@ nv_pci_count_devices(void)
                 pci_dev->subsystem_device,
                 NV_TRUE /* print_legacy_warning */))
         {
+            check_for_bound_driver(pci_dev);
             count++;
         }
         pci_dev = pci_get_class(PCI_CLASS_DISPLAY_VGA << 8, pci_dev);
@@ -925,6 +963,7 @@ nv_pci_count_devices(void)
                 pci_dev->subsystem_device,
                 NV_TRUE /* print_legacy_warning */))
         {
+            check_for_bound_driver(pci_dev);
             count++;
         }
         pci_dev = pci_get_class(PCI_CLASS_DISPLAY_3D << 8, pci_dev);
